/* A cilk compiler driver that does not depend on gcc. */
/* use it like you would gcc.
 * How cilk gets compiled:
 *    foo.cilk  -> foo.cilki  run cpp (use gcc -E)  [call this cpp1]
 *    foo.cilki -> foo.cilkc  run cilk2c
 *    foo.cilkc -> foo.i      run cpp again (use gcc -E)
 *    *.i *.s *.o *.a -> foo.s *.o *.a      run gcc
 * You may link  in c functions too
 *    foo.c     -> foo.S      run gcc
 *  note that gcc may skip producing foo.S or foo.o if you invoke it properly. 
'*  We let gcc take care of everything after converting foo.cilk to foo.i
 *    
 * Arguments
 *   -x <language>            like gcc -x (accept "cilk" or anything that gcc accepts, e.g., "c")
 *   foo.cilk                 process with gcc -E, then, cilk2c, then gcc to produce a file .o, then the gcc link phase (if linking)
 *   foo.c (or other files)   process with gcc
 *   -o <fname>               name of output file
 *   -E0                      stop after preprocessing the cilk (to get .cilki)
 *   -E1                      stop after transforming cilk to c (to get .cilkc)
 *   -E                       stop when all preprocessing is done (when you get a .i file)
 *   -S                       stop when assembly files are produced
 *   -c                       stop when .o files are produced
 *   -D* -A* -include -I      give to cpp
 *   -W*                      give to gcc invocations
 *   -f*                      give to gcc invocations
 *   -save-temps              save the temporary files (also call them foo.i and foo.s instead of /tmp/cXF*xc.i)
 *   -cilk-debug
 *   -cilk-profile
 *   -cilk-critical-path
 *   -cilk-hint
 *
 * These restrictions come from gcc:
 *  There is no provision for filenames that start with -.
 *  There is no way to treat a file name named foo.c as a .o file or a .a file. 
 *
 * other gcc-style arguments to be added later
 */
static const char *ident __attribute__((__unused__))
     = "$HeadURL: https://bradley.lcs.mit.edu/svn/repos/cilk/5.4.2.2/support/cilkcB.c $ $LastChangedBy: sukhaj $ $Rev: 1708 $ $Date: 2003-10-27 10:19:53 -0500 (Mon, 27 Oct 2003) $";

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>

#include "make-temp-file.h"

/* This macro hacking is what is needed to get RUNTIME_DIR to be expanded and then stringified */
#define STR(S) #S
#define XSTR(S) STR(S)
#define CILK_H_LOCATION XSTR(CILKHEADER_DIR)"/cilk.h"
#define CILK2C_LOCATION XSTR(CILK2C_DIR)"/cilk2c"

static int match (const char *s1, const char *s2);
static int tailmatch (const char *s1, const char *tail);
static void barf (const char *format, ...) __attribute__((noreturn)) __attribute__((format (printf, 1, 2)));

static int match (const char *s1, const char *s2) {
    return strcmp(s1,s2)==0;
}
static int tailmatch (const char *s1, const char *tail) {
    int l1=strlen(s1);
    int l2=strlen(tail);
    if (l1<l2) return 0;
    return strcmp(s1+l1-l2, tail)==0;
}
static int headmatch (const char *s1, const char *head) {
    int l1=strlen(s1);
    int l2=strlen(head);
    if (l1<l2) return 0;
    return strncmp(s1, head, l2)==0;
}

static void barf (const char *format, ...) {
    va_list ap;
    va_start(ap, format);
    fprintf(stderr, "BARF: ");
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, "\n");
    exit(1);
}

enum stope { STOP_AT_CILKI, /* stop after running cpp on cilk sources*/
	     STOP_AT_CILKC, /* stop after running cilk2c */
	     STOP_AT_I,     /* stop after cpp on cilkc and cpp on .c files (-E to gcc) */
	     STOP_AT_S,     /* stop after assembly. */
	     STOP_AT_O,     /* stop after generating the .o file (i.e., pass -c to gcc */
	     STOP_AT_M,     /* print dependencies only. */
	     STOP_AT_END } stop_at = STOP_AT_END;


typedef struct stringlist *STRINGLIST;
struct stringlist {
    const char *first;
    STRINGLIST  rest;
};

static void pushstring (STRINGLIST *sl, const char *s) {
    STRINGLIST n=malloc(sizeof(*n));
    assert(n);
    n->first=s;
    n->rest =*sl;
    *sl = n;
}

typedef struct tmpfilelist *TMPFILELIST;
struct tmpfilelist {
    const char *temp_filename;
    const char *org_filename;
    TMPFILELIST rest;
};

static const char * pushtempfile (TMPFILELIST *sl, const char *temp_filename, const char *org_filename) {
    TMPFILELIST n=malloc(sizeof(*n));
    assert(n);
    n->temp_filename = temp_filename;
    n->org_filename = org_filename;
    n->rest =*sl;
    *sl = n;
    return temp_filename;
}

typedef struct my_argv {
    unsigned int argc, limit;
    char const**argv;
} *MY_ARGV;

static MY_ARGV make_my_argv(void) {
    MY_ARGV result = malloc(sizeof(*result));
    enum { INITIAL_LIMIT = 5 };
    result->argc=0;
    result->limit = INITIAL_LIMIT;
    result->argv = malloc(INITIAL_LIMIT*sizeof(char*));
    result->argv[0]=0; // Always leave a null at the end.
    //printf("size=%d argv=%p\n", INITIAL_LIMIT*sizeof(char*), result->argv);
    return result;
}
static void free_my_argv(MY_ARGV ma) {
    free(ma->argv);
    free(ma);
}
/*
static void reset_my_argv(MY_ARGV ma) {
    ma->argc=0;
    ma->argv[0]=0;
}
*/
static void add_arg_to_my_argv(MY_ARGV ma, const char *s) {
    if (ma->argc+1>=ma->limit) { /* Always leave one extra for the NULL at the end. */
	ma->limit*=2;
	ma->argv=realloc(ma->argv, ma->limit*sizeof(char*));
	//printf("new size is %d, ptr is %p\n", ma->limit*sizeof(char*), ma->argv); 
    }
    ma->argv[ma->argc++] = s;
    ma->argv[ma->argc] = 0;
    assert(ma->argc<=ma->limit);
}
static void print_my_argv(FILE *f, MY_ARGV ma) {
    unsigned int i;
    for (i=0; i<ma->argc; i++) {
	if (i>0) fprintf(f, " ");
	fprintf(f, "%s", ma->argv[i]);
    }
    assert(ma->argv[ma->argc]==0);
    assert(ma->argc+1<=ma->limit);
    fprintf(f, "\n");
}

static void add_list_to_my_argv_backwards(MY_ARGV ma, STRINGLIST l) {
    if (l==0) return;
    add_list_to_my_argv_backwards(ma, l->rest);
    add_arg_to_my_argv(ma, l->first);
}

static int my_system(MY_ARGV ma) {
    pid_t child;
    int status;
    if ((child=fork())==0) {
	execvp(ma->argv[0], (char **)ma->argv);
	abort();
    }
    waitpid(child, &status, 0);
    //printf("status=%d exited=%d\n", status, WIFEXITED(status));
    return !(WIFEXITED(status) && WEXITSTATUS(status)==0);
}

/* These are filled in by parse_args: */

static const char *output_name;
static int cilk_profile=0, cilk_debug=0, cilk_critical_path=0, cilk_hint=0;
static int verbose=0;
static int save_temps = 0; /* true if we saw -save-temps */
static const char *stop_at_m_arg=0;
static STRINGLIST cpp1_args=0;  /* preprocessor args only for the pass before cilk2c. */ 
static STRINGLIST preprocessor_args=0;
static STRINGLIST cc_args=0;
static STRINGLIST linker_args=0;
/* filenames */
static STRINGLIST cilk_files=0, cilki_files=0, cilkc_files=0;
static STRINGLIST c_files=0, i_files=0, ii_files=0, m_files=0, h_files=0, cpp_files=0, s_files=0, S_files=0, other_files=0;
static unsigned int cilk_files_count=0, cilki_files_count=0, cilkc_files_count=0;
static unsigned int c_files_count=0, i_files_count=0, ii_files_count=0, m_files_count=0, h_files_count=0, cpp_files_count=0, s_files_count=0, S_files_count=0, other_files_count=0;

static TMPFILELIST cilki_tempfiles=0, cilkc_tempfiles=0, i_tempfiles=0;

#define VERBOSE(level,expr) ({ if (verbose>=(level)) (expr); })

/* new_suffix includes the '.' */
static const char* replace_suffix (const char *filename, const char *new_suffix) {
    char *enda = strrchr(filename, '.');
    if (enda==0) {
	int flen = strlen(filename);
	int len = flen + strlen(new_suffix) + 1;
	char *result = malloc(len);
	strcpy(result, filename); // This won't overflow.
	strcpy(result+flen, new_suffix); // Nor will this */
	return result;
    } else {
	int flen = enda-filename;
	int len = flen + strlen(new_suffix) + 1;
	char *result = malloc(len);
	strncpy(result, filename, flen);
	strcpy(result+flen, new_suffix);
	return result;
    }}


static TMPFILELIST all_tempfiles;

/* suffix includes the `.` */
static const char * newtempfile (TMPFILELIST *sl, const char *suffix, const char *org_filename) {
    const char *fname;
    if (save_temps) {
	fname = replace_suffix(org_filename, suffix);
    } else {
	fname = make_temp_file(suffix);
    }
    pushtempfile(&all_tempfiles, fname, org_filename);
    pushtempfile(sl, fname, org_filename);
    return fname;
}

static int parse_args (int argc, const char **argv) {
    STRINGLIST    *files_list=0;
    unsigned int  *files_count=0;
    const char *last_stop_at = 0;
    void note_stop_at (enum stope sa) {
	if (stop_at!=STOP_AT_END) barf("%s incompatible with %s", last_stop_at,*argv);
	stop_at = sa;
    }
    argc--; argv++;
    while (argc>0) {
	if (match(*argv, "-E0"))      note_stop_at(STOP_AT_CILKI);
	else if (match(*argv, "-E1")) note_stop_at(STOP_AT_CILKC);
	else if (match(*argv, "-E"))  note_stop_at(STOP_AT_I);
	else if (match(*argv, "-c"))  note_stop_at(STOP_AT_O);
	else if (match(*argv, "-S"))  note_stop_at(STOP_AT_S);
	else if (match(*argv, "-v") || match(*argv, "-v2")) {
	    verbose=2;
	} else if (match(*argv,"-v1")) {
	    verbose=1;
	} else if (match(*argv, "-cilk-profile")) {
	    cilk_profile=1;
	} else if (match(*argv, "-cilk-debug")) {
	    cilk_debug=1;
	} else if (match(*argv, "-cilk-critical-path")) {
	    cilk_critical_path=1;
	} else if (match(*argv, "-cilk-hint")) {
	    cilk_hint=1;
	} else if (match(*argv, "-save-temps")) {
	    save_temps=1;
	} else if (match(*argv, "-M") ||
		   match(*argv, "-MM")) {
	    // These only to cpp, and they stop after doing the preprocessing
	    note_stop_at(STOP_AT_M);
	    stop_at_m_arg=*argv;
	} else if (match(*argv, "-I")) {
	    /* Pass this and the next argument to the preprocessor. */
	    if (argc==1) barf("Missing arg to %s", *argv);
	    pushstring(&preprocessor_args, *argv);
	    argc--; argv++;
	    pushstring(&preprocessor_args, *argv);
	} else if (headmatch(*argv,"-D") ||
		   headmatch(*argv,"-A")  ||
		   headmatch(*argv,"-include")  ||
		   headmatch(*argv,"-I") ||
		   headmatch(*argv,"-Wp,")) {
	    /* Pass these to the preprocessor. */
	    pushstring(&preprocessor_args, *argv);
	} else if (match(*argv, "-g") ||
		   headmatch(*argv, "-O") ||
		   headmatch(*argv, "-Wl") ||
		   headmatch(*argv, "-m")) {
	    pushstring(&cc_args, *argv);
	    pushstring(&linker_args, *argv);
	} else if (match(*argv, "-l") ||
		   match(*argv, "-L")
		   ) {
	    /* Pass these to gcc only. */
	    if (argc==1) barf("Missing arg to %s", *argv);
	    pushstring(&linker_args, *argv);
	    argc--; argv++;
	    pushstring(&linker_args, *argv);
	} else if (headmatch(*argv, "-l") || /* if it was "-l foo.a" then the previous case matched, otherwise it is "-lfoo.a" here. */
		   headmatch(*argv, "-L") ||
		   match(*argv, "-static")
		   ) {
	    /* Pass these to gcc only. */
	    pushstring(&linker_args, *argv);
	} else if (match(*argv, "-o")) {
	    if (argc==1) barf("Missing arg to -x");
	    argc--; argv++;
	    if (output_name!=0) barf("Two output files specified");
	    output_name=*argv;
	} else if (match(*argv, "-x")) {
	    if (argc==1) barf("Missing arg to -x");
	    argc--; argv++;
	    if (match(*argv, "cilk"))               { files_list = &cilk_files;   files_count = &cilk_files_count; }
	    else if (match(*argv, "cilki"))         { files_list = &cilki_files;  files_count = &cilki_files_count;}
	    else if (match(*argv, "cilkc"))         { files_list = &cilkc_files;  files_count = &cilkc_files_count;}
	    else if (match(*argv, "c"))             { files_list = &c_files;      files_count = &c_files_count;}
	    else if (match(*argv, "objective-c"))   { files_list = &m_files;      files_count = &m_files_count;}
	    else if (match(*argv, "c++"))           { files_list = &cpp_files;    files_count = &cpp_files_count;}
	    else if (match(*argv, "c-header"))      { files_list = &h_files;      files_count = &h_files_count;}
	    else if (match(*argv, "cpp-output"))    { files_list = &i_files;      files_count = &i_files_count;}
	    else if (match(*argv, "c++cpp-output")) { files_list = &ii_files;     files_count = &ii_files_count;}
	    else if (match(*argv, "assembler"))     { files_list = &s_files;      files_count = &s_files_count;}
	    else if (match(*argv, "assembler-with-cpp")) { files_list = &S_files; files_count = &S_files_count; }
	    else if (match(*argv, "none"))          { files_list = 0;             files_count = 0; }
	    else
		barf("What is this language?  %s", *argv);
	} else if (**argv!='-') { /* It is a filename */
	    if (files_list) {
		pushstring(files_list, *argv); (*files_count)++;
	    } else {
		if      (tailmatch(*argv, ".cilk"))  { pushstring(&cilk_files,  *argv); cilk_files_count++; }
		else if (tailmatch(*argv, ".cilki")) { pushstring(&cilki_files, *argv); cilki_files_count++; }
		else if (tailmatch(*argv, ".cilkc")) { pushstring(&cilkc_files, *argv); cilkc_files_count++; }
		else if (tailmatch(*argv, ".c"))     { pushstring(&c_files,     *argv); c_files_count++; }
		else if (tailmatch(*argv, ".i"))     { pushstring(&i_files,     *argv); c_files_count++; }
		else if (tailmatch(*argv, ".ii"))    { pushstring(&ii_files,    *argv); ii_files_count++; }
		else if (tailmatch(*argv, ".m"))     { pushstring(&m_files,     *argv); m_files_count++; }
		else if (tailmatch(*argv, ".h"))     { pushstring(&h_files,     *argv); h_files_count++; }
		else if (tailmatch(*argv, ".cc") ||
			 tailmatch(*argv, ".cxx") ||
			 tailmatch(*argv, ".cpp") ||
			 tailmatch(*argv, ".C"))     { pushstring(&cpp_files,   *argv); cpp_files_count++; }
		else if (tailmatch(*argv, ".s"))     { pushstring(&s_files,     *argv); s_files_count++; }
		else if (tailmatch(*argv, ".S"))     { pushstring(&S_files,     *argv); S_files_count++; }
		else                                 { pushstring(&other_files, *argv); other_files_count++; }
	    }
	} else {
	    /* I don't know what it is. */
	    barf("What is this? %s", *argv);
	}
	argc--; argv++;
    }
    return 0;
}
    /* If you specify output_name but more than one target, they will all write to the same output. */

/* Run cpp1, the first preprocessing step. */
static void compile_cilk_file (const char *fname) {
    int r;
    MY_ARGV ma = make_my_argv();
    add_arg_to_my_argv(ma, "gcc");
    add_arg_to_my_argv(ma, "-E");
    add_arg_to_my_argv(ma, "-include");
    add_arg_to_my_argv(ma, CILK_H_LOCATION);
    add_arg_to_my_argv(ma, "-x");
    add_arg_to_my_argv(ma, "c");
    add_arg_to_my_argv(ma, fname);
    add_arg_to_my_argv(ma, "-o");
    if (stop_at == STOP_AT_CILKI) {
	if (output_name) {
	    add_arg_to_my_argv(ma, output_name);
	} else {
	    add_arg_to_my_argv(ma, replace_suffix(fname, ".cilki"));
	}
    } else {
	add_arg_to_my_argv(ma, newtempfile(&cilki_tempfiles, ".cilki", fname));
	cilki_files_count++;
    }
    add_list_to_my_argv_backwards(ma, cpp1_args);
    add_list_to_my_argv_backwards(ma, preprocessor_args);
    if (verbose==2) {
	add_arg_to_my_argv(ma, "-v");
    }
    VERBOSE(1, print_my_argv(stderr, ma));
    r = my_system(ma); 
    if (r!=0) {
	barf("failed running cpp1");
    }
    free_my_argv(ma);
}

/* Run cilk2c */
static void compile_cilki_file (const char *fname, const char *original_fname) {
    int r;
    MY_ARGV ma = make_my_argv();
    add_arg_to_my_argv(ma, CILK2C_LOCATION);
    add_arg_to_my_argv(ma, fname);
    if (cilk_hint)
	add_arg_to_my_argv(ma, "-hint");
    add_arg_to_my_argv(ma, "-o");
    if (stop_at == STOP_AT_CILKC) {
	if (output_name) {
	    add_arg_to_my_argv(ma, output_name);
	} else {
	    add_arg_to_my_argv(ma, replace_suffix(original_fname, ".cilkc"));
	}
    } else {
	add_arg_to_my_argv(ma, newtempfile(&cilkc_tempfiles, ".cilkc", original_fname));
	cilkc_files_count++;
    }
    if (verbose==2) {
	add_arg_to_my_argv(ma, "-v");
    }
    VERBOSE(1,print_my_argv(stderr, ma));
    r = my_system(ma); 
    if (r!=0) {
	barf("failed running cilk2c");
    }
    free_my_argv(ma);
}

/* Run cpp2, the second preprocesing step */
static void compile_cilkc_file (const char *fname, const char *original_fname) {
    int r;
    MY_ARGV ma = make_my_argv();
    add_arg_to_my_argv(ma, "gcc");
    add_arg_to_my_argv(ma, "-E");
    add_arg_to_my_argv(ma, "-x");
    add_arg_to_my_argv(ma, "c");
    add_arg_to_my_argv(ma, "-I" XSTR(CILKHEADER_DIR));
    add_arg_to_my_argv(ma, "-D_REENTRANT");
    add_arg_to_my_argv(ma, fname);
    add_arg_to_my_argv(ma, "-o");
    if (stop_at==STOP_AT_I) {
	if (output_name) {
	    add_arg_to_my_argv(ma, output_name);
	} else {
	    add_arg_to_my_argv(ma, replace_suffix(original_fname, ".i"));
	}
    } else {
	add_arg_to_my_argv(ma, newtempfile(&i_tempfiles, ".i", original_fname));
	i_files_count++;
    }
    if (verbose==2) {
	add_arg_to_my_argv(ma, "-v");
    }
    VERBOSE(1, print_my_argv(stderr, ma));
    r = my_system(ma); 
    if (r!=0) {
	barf("failed running cpp2");
    }
    free_my_argv(ma);
}

int do_operations(void) {
    VERBOSE(1, fprintf(stderr, "cilkc $Rev: 1708 $ $Date: 2003-10-27 10:19:53 -0500 (Mon, 27 Oct 2003) $\n"));
    pushstring(&cpp1_args, "-I" XSTR(CILKHEADER_DIR) );
    pushstring(&cpp1_args, "-D__CILK__");
    pushstring(&cpp1_args, "-D__CILK2C__");
    pushstring(&cpp1_args, "-D__REENTRANT");
    if (cilk_debug)   pushstring(&cpp1_args, "-DCILK_DEBUG=1");
    if (cilk_profile) {
	pushstring(&cpp1_args, "-DCILK_STATS=1");
	pushstring(&cpp1_args, "-DCILK_TIMING=1");
    }
    if (cilk_critical_path) {
	if (!cilk_profile) {
	    barf("-cilk-critical-path requires -cilk-profile");
	}
	pushstring(&cpp1_args, "-DCILK_CRITICAL_PATH=1");
    }
    if (cilk_debug) {
	pushstring(&cpp1_args, "-DCILK_DEBUG=1");
    }

    /* Now do the commands.*/
    if (stop_at==STOP_AT_M) {
	if (cilk_files) {
	    /* Create dependencies for cilk programs. */
	    int r;
	    MY_ARGV ma = make_my_argv();
	    add_arg_to_my_argv(ma, "gcc");
	    add_arg_to_my_argv(ma, "-M");
	    add_arg_to_my_argv(ma, "-x");
	    add_arg_to_my_argv(ma, "c");
	    add_list_to_my_argv_backwards(ma, cilk_files);
	    add_list_to_my_argv_backwards(ma, cpp1_args);
	    add_list_to_my_argv_backwards(ma, preprocessor_args);
	    if (verbose==2) {
		add_arg_to_my_argv(ma, "-v");
	    }
	    VERBOSE(1, print_my_argv(stderr, ma));
	    r = my_system(ma); 
	    if (r!=0) {
		barf("creating dependencies for cilk files");
	    }
	    free_my_argv(ma);
	}
	if (c_files || cpp_files || S_files || m_files) {
	    /* Create dependencies for everything else. */
	    int r;
	    MY_ARGV ma = make_my_argv();
	    add_arg_to_my_argv(ma, "gcc");
	    add_arg_to_my_argv(ma, "-M");
	    add_list_to_my_argv_backwards(ma, preprocessor_args);
	    if (c_files) {
		add_arg_to_my_argv(ma, "-x");
		add_arg_to_my_argv(ma, "c");
		add_list_to_my_argv_backwards(ma, c_files);
	    }
	    if (cpp_files) {
		add_arg_to_my_argv(ma, "-x");
		add_arg_to_my_argv(ma, "c++");
		add_list_to_my_argv_backwards(ma, cpp_files);
	    }
	    if (S_files) {
		add_arg_to_my_argv(ma, "-x");
		add_arg_to_my_argv(ma, "assembler-with-cpp");
		add_list_to_my_argv_backwards(ma, S_files);
	    }
	    if (m_files) {
		add_arg_to_my_argv(ma, "-x");
		add_arg_to_my_argv(ma, "objective-c");
		add_list_to_my_argv_backwards(ma, m_files);
	    }
	    if (verbose==2) {
		add_arg_to_my_argv(ma, "-v");
	    }
	    VERBOSE(1, print_my_argv(stderr, ma));
	    r = my_system(ma); 
	    if (r!=0) {
		barf("creating dependencies for gcc files");
	    }
	    free_my_argv(ma);
	}
	return 0; /* -M finished OK */
    }
    /* run cpp1 (the preprocessor before cilk2c */ 
    {
	STRINGLIST h;
	if ((stop_at == STOP_AT_CILKI) && cilk_files_count>1 && output_name)
	    barf("You may not specify an output file with -E0 and more than one cilk input file");

	for (h=cilk_files; h; h=h->rest) {
	    compile_cilk_file(h->first);
	}
    }
    if (stop_at==STOP_AT_CILKI) return 0;
    /* Now run cilk2c, converting .cilki -> .cilkc */
    {
	if ((stop_at==STOP_AT_CILKC) && cilki_files_count>1 && output_name) 
	    barf("You may not specify an output file with -E1 and multiple input files");

	{
	    STRINGLIST h;
	    for(h=cilki_files; h; h=h->rest)
		compile_cilki_file(h->first, h->first);
	}
	{
	    TMPFILELIST h;
	    for(h=cilki_tempfiles; h; h=h->rest)
		compile_cilki_file(h->temp_filename, h->org_filename);
	}
    }

    if (stop_at==STOP_AT_CILKC) return 0;
    /* Now run cpp2 converting .cilkc -> .i */
    {
	if ((stop_at==STOP_AT_I) && (cilkc_files_count+c_files_count)>1 && output_name)
	    barf("You may not specify an output file with -E and multiple input files");
	    
	{
	    STRINGLIST h;
	    for(h=cilkc_files; h; h=h->rest)
		compile_cilkc_file(h->first, h->first);
	}
	{
	    TMPFILELIST h;
	    for(h=cilkc_tempfiles; h; h=h->rest)
		compile_cilkc_file(h->temp_filename, h->org_filename);
	}
    }
    if (stop_at==STOP_AT_I) return 0;
    if (stop_at==STOP_AT_O || stop_at==STOP_AT_S) {
	/* We must compile the I files separately, because their temp file names are not good enough. */
	TMPFILELIST h;
	if (output_name && i_files_count>1)
	    barf("You may not specify an output file with -S or -c and multiple input files");
	for (h=i_tempfiles; h; h=h->rest) {
	    int r;
	    MY_ARGV ma = make_my_argv();
	    add_arg_to_my_argv(ma, "gcc");

	    if (stop_at==STOP_AT_S)      add_arg_to_my_argv(ma, "-S");
	    else if (stop_at==STOP_AT_O) add_arg_to_my_argv(ma, "-c");
	    else barf("Internal compiler error %s:%d", __FILE__, __LINE__);

	    add_arg_to_my_argv(ma, h->temp_filename);
	    add_arg_to_my_argv(ma, "-o");
	    if (output_name) {
		add_arg_to_my_argv(ma, output_name);
	    } else {
		add_arg_to_my_argv(ma, replace_suffix(h->org_filename, ".o"));
	    }
	    add_list_to_my_argv_backwards(ma, cc_args);
	    if (verbose==2) {
		add_arg_to_my_argv(ma, "-v");
	    }
	    VERBOSE(1, print_my_argv(stderr, ma));
	    r = my_system(ma);
	    if (r!=0) {
		fprintf(stderr, "gcc failed\n");
		return 1;
	    }
	    free_my_argv(ma);
	}
    }
    /* Now we can run gcc on the .i files and everything else */
    if ((stop_at==STOP_AT_O && (c_files || i_files || ii_files || m_files || cpp_files || s_files || S_files))
	|| (stop_at==STOP_AT_I && (c_files || m_files || S_files))
	|| (stop_at==STOP_AT_END)) {
	int r;
	STRINGLIST h;
	MY_ARGV ma = make_my_argv();
	add_arg_to_my_argv(ma, "gcc");

	if (stop_at==STOP_AT_S)      add_arg_to_my_argv(ma, "-S");
	else if (stop_at==STOP_AT_O) add_arg_to_my_argv(ma, "-c");

	add_list_to_my_argv_backwards(ma, cc_args);

	if (h_files) {
	    for (h=h_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (c_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "c");
	    add_list_to_my_argv_backwards(ma, preprocessor_args); // When compiling C files, we need the preprocessor arguments. */
	    for (h=c_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (i_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "cpp-output");
	    for (h=i_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (stop_at==STOP_AT_END) {
	    /* If we are not doing -S or -c, then we need to include the temp .i files here. */
	    TMPFILELIST tl;
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "cpp-output");
	    for (tl=i_tempfiles; tl; tl=tl->rest) {
		add_arg_to_my_argv(ma, tl->temp_filename);
	    }
	}
	if (ii_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "c++-pp-output");
	    for (h=i_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (m_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "objective-c");
	    for (h=m_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (cpp_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "c++");
	    if (!c_files)
		/* When compiling C++ files, we need the preprocessor arguments. */
		/* But don't add them again if we are compiling C files, since we added them above. */
		add_list_to_my_argv_backwards(ma, preprocessor_args);
	    for (h=cpp_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (s_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "assembler");
	    for (h=s_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (S_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "assembler-with-cpp");
	    if (!c_files && ! cpp_files) {
		/* When compiling C++ files, we need the preprocessor arguments. */
		/* But don't add them again if we added them above for the C or CPP files. */
		add_list_to_my_argv_backwards(ma, preprocessor_args);
	    }
	    for (h=s_files; h; h=h->rest)  
		add_arg_to_my_argv(ma, h->first);
	}
	if (other_files) {
	    add_arg_to_my_argv(ma, "-x"); add_arg_to_my_argv(ma, "none");
	    add_list_to_my_argv_backwards(ma, other_files);
	}

	if (stop_at==STOP_AT_END) {

	    add_list_to_my_argv_backwards(ma, linker_args);

	    add_arg_to_my_argv(ma, "-L" XSTR(LIBS_DIR));
	    add_arg_to_my_argv(ma, "-L" XSTR(LIBS2_DIR));
	    if (cilk_debug) {
		if (cilk_profile) { add_arg_to_my_argv(ma, "-lcilkrt0gp"); add_arg_to_my_argv(ma, "-lcilk.g.p"); }
		else              { add_arg_to_my_argv(ma, "-lcilkrt0g");  add_arg_to_my_argv(ma, "-lcilk.g");   }
	    } else {
		if (cilk_profile) { add_arg_to_my_argv(ma, "-lcilkrt0p");  add_arg_to_my_argv(ma, "-lcilk.p");   }
		else              { add_arg_to_my_argv(ma, "-lcilkrt0");   add_arg_to_my_argv(ma, "-lcilk");     }
	    }
	    // -R followed by a pathname is treated as -rpath in the gnu linker,
	    // and it works in solaris.
	    // However -R doesn't work on irix, and -rpath does.
	    // So RPATH_NAME is either -rpath (ok for linux or sgi) or -R (ok for linux or solaris) 
	    if (strcmp(XSTR(HAVE_RPATH), "yes")==0)
		add_arg_to_my_argv(ma, "-Wl," XSTR(HAVE_RPATHARG) "," XSTR(LIBS_DIR));
	}
	if (output_name) {
	    add_arg_to_my_argv(ma, "-o");
	    add_arg_to_my_argv(ma, output_name);
	}
	if (verbose==2) {
	    add_arg_to_my_argv(ma, "-v");
	}
   
	VERBOSE(1, print_my_argv(stderr, ma));
	r = my_system(ma);
	if (r!=0) {
	    fprintf(stderr, "gcc failed\n");
	    return 1;
	}
	free_my_argv(ma);
    }
    return 0;
}

int main (int argc, const char *argv[]) {
    int r = parse_args(argc, argv);
    if (r==0) {
	r = do_operations();
    }
    if (!save_temps) {
	TMPFILELIST tl;
	for (tl=all_tempfiles; tl; tl=tl->rest) {
	    unlink(tl->temp_filename);
	}
    }
    return r;
}
